import json
from copy import deepcopy

from elasticsearch import VERSION as ES_VERSION
from invenio_records_rest.serializers import record_responsify, search_responsify
from oarepo_validate import JSONSerializer

lt_es7 = ES_VERSION[0] < 7


# TODO: přesunout, pravděpodobně do nr-common
class NrJSONSerializer(JSONSerializer):
    def serialize_search(self, pid_fetcher, search_result, links=None, item_links_factory=None,
                         **kwargs):
        """Serialize a search result.

                :param pid_fetcher: Persistent identifier fetcher.
                :param search_result: Elasticsearch search result.
                :param links: Dictionary of links to add to response.
                """
        search_result = self.post_process_search_result(search_result)
        total = search_result['hits']['total'] if lt_es7 else \
            search_result['hits']['total']['value']
        return json.dumps(dict(
            hits=dict(
                hits=[self.transform_search_hit(
                    pid_fetcher(hit['_id'], hit['_source']),
                    hit,
                    links_factory=item_links_factory,
                    **kwargs
                ) for hit in search_result['hits']['hits']],
                total=total,
            ),
            links=links or {},
            aggregations=search_result.get('aggregations', dict()),
        ), **self._format_args())

    @staticmethod
    def post_process_search_result(search_result):
        access_rights_container = search_result.get("aggregations", {}).get("nested#accessRights", None)
        if access_rights_container:
            access_rights = access_rights_container.get("sterms#inner_facet", None)
            if access_rights:
                buckets = access_rights.get("buckets", [])
                open_access: int = 0
                close_access: int = 0
                for bucket in buckets:
                    if bucket["key"] == "open access":
                        open_access += bucket["doc_count"]
                    else:
                        close_access += bucket["doc_count"]
                new_buckets = [
                    {"key": 1, "key_as_string": "true", "doc_count": open_access},
                    {"key": 0, "key_as_string": "false", "doc_count": close_access}
                ]
                new_buckets = [bucket for bucket in new_buckets if bucket["doc_count"] > 0]
                is_open_access = deepcopy(access_rights)
                is_open_access["buckets"] = new_buckets
                access_rights_container["sterms#inner_facet"] = is_open_access
        return search_result


json_serializer = NrJSONSerializer(replace_refs=False)

json_response = record_responsify(json_serializer, 'application/json')
json_search = search_responsify(json_serializer, 'application/json')



#
# Copyright (c) 2019 UCT Prague.
#
# fetchers.py is part of CIS KROKD repository
# (see https://cis-git.vscht.cz/cis/cis-repo-invenio/cis-krokd-repository).
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
#
import functools

from invenio_pidstore.fetchers import FetchedPID
from nr_events.record import EventBaseRecord
from nr_generic.fetchers import nr_id_generic_fetcher
from nr_events.fetchers import nr_events_id_fetcher
from nr_generic.record import CommonBaseRecord
from nr_nresults.record import NResultBaseRecord
from nr_theses.fetchers import nr_theses_id_fetcher
from nr_nresults.fetchers import nr_nresults_id_fetcher
from nr_theses.record import ThesisBaseRecord
from nr_datasets.record import DatasetBaseRecord
from nr_datasets.fetchers import nr_datasets_id_fetcher


@functools.lru_cache(maxsize=1)
def prepare_schemas():
    CommonBaseRecord._prepare_schemas()
    EventBaseRecord._prepare_schemas()
    ThesisBaseRecord._prepare_schemas()
    NResultBaseRecord._prepare_schemas()
    DatasetBaseRecord._prepare_schemas()


def nr_all_id_fetcher(record_uuid, data):
    """Fetch an object restoration PID.

        :param record_uuid: Record UUID.
        :param data: Record content.
        :returns: A :class:`invenio_pidstore.fetchers.FetchedPID` that contains
            data['did'] as pid_value.
        """
    if '$schema' not in data:
        raise ValueError('Return _schema in search results')

    schema = data['$schema']

    prepare_schemas()

    if schema == CommonBaseRecord.PREFERRED_SCHEMA:
        fetched_pid = nr_id_generic_fetcher(record_uuid, data)
    elif schema == EventBaseRecord.PREFERRED_SCHEMA:
        fetched_pid = nr_events_id_fetcher(record_uuid, data)
    elif schema == ThesisBaseRecord.PREFERRED_SCHEMA:
        fetched_pid = nr_theses_id_fetcher(record_uuid, data)
    elif schema == NResultBaseRecord.PREFERRED_SCHEMA:
        fetched_pid = nr_nresults_id_fetcher(record_uuid, data)
    elif schema == DatasetBaseRecord.PREFERRED_SCHEMA:
        fetched_pid = nr_datasets_id_fetcher(record_uuid, data)
    else:
        raise ValueError(f'Unknown conversion of schema "{schema}" to pid type')

    if data.get('oarepo:draft'):
        return FetchedPID(
            provider=fetched_pid.provider,
            pid_type="d" + fetched_pid.pid_type,
            pid_value=fetched_pid.pid_value,
        )
    else:
        return fetched_pid



from functools import partial

from invenio_records_rest.utils import deny_all, check_elasticsearch
from nr_common.links import nr_links_factory
from nr_common.search import community_search_factory
from nr_generic.config import FACETS, CURATOR_FACETS, FILTERS, CURATOR_FILTERS
from oarepo_communities.links import community_record_links_factory
from oarepo_records_draft import DRAFT_IMPORTANT_FACETS, DRAFT_IMPORTANT_FILTERS
from oarepo_ui.facets import translate_facets

from nr_all.constants import all_index_name
from nr_all.record import AllNrRecord
from nr_all.search import AllRecordsSearch, AllRecordsDraftSearch, AllRecordsPublishedSearch

RECORDS_REST_ENDPOINTS = {
    # readonly url for both endpoints, does not have item route
    # as it is accessed from the endpoints above
    'all': dict(
        pid_type='nrall',
        pid_minter='nr_all',
        pid_fetcher='nr_all',
        default_endpoint_prefix=True,
        search_class=AllRecordsSearch,
        record_class=AllNrRecord,
        search_index=all_index_name,
        search_serializers={
            'application/json': 'nr_all.serializers:json_search',
        },
        list_route='/all/',
        default_media_type='application/json',
        max_result_window=1000000,
        search_factory_imp=community_search_factory,
        links_factory_imp=partial(community_record_links_factory,
                                  original_links_factory=nr_links_factory),

        # not used really
        item_route='/all/not-used-but-must-be-present',
        create_permission_factory_imp=deny_all,
        delete_permission_factory_imp=deny_all,
        update_permission_factory_imp=deny_all,
        read_permission_factory_imp=check_elasticsearch,
        record_serializers={
            'application/json': 'oarepo_validate:json_response',
        },
        # search_factory_imp='restoration.objects.search:objects_search_factory'
        # default search_factory_imp: invenio_records_rest.query.default_search_factory
    ),
    'draft': dict(
        pid_type='nrall-draft',
        pid_minter='nr_all',
        pid_fetcher='nr_all',
        default_endpoint_prefix=True,
        search_class=AllRecordsDraftSearch,
        record_class=AllNrRecord,
        search_index=all_index_name,
        search_serializers={
            'application/json': 'nr_all.serializers:json_search',
        },
        list_route='/draft/',
        default_media_type='application/json',
        max_result_window=1000000,
        search_factory_imp=community_search_factory,
        links_factory_imp=partial(community_record_links_factory,
                                  original_links_factory=nr_links_factory),

        # not used really
        item_route='/all/not-used-but-must-be-present',
        create_permission_factory_imp=deny_all,
        delete_permission_factory_imp=deny_all,
        update_permission_factory_imp=deny_all,
        read_permission_factory_imp=check_elasticsearch,
        record_serializers={
            'application/json': 'oarepo_validate:json_response',
        },
        # search_factory_imp='restoration.objects.search:objects_search_factory'
        # default search_factory_imp: invenio_records_rest.query.default_search_factory
    ),
    'published': dict(
        pid_type='nrall-published',
        pid_minter='nr_all',
        pid_fetcher='nr_all',
        default_endpoint_prefix=True,
        search_class=AllRecordsPublishedSearch,
        record_class=AllNrRecord,
        search_index=all_index_name,
        search_serializers={
            'application/json': 'nr_all.serializers:json_search',
        },
        list_route='/published/',
        default_media_type='application/json',
        max_result_window=1000000,
        search_factory_imp=community_search_factory,
        links_factory_imp=partial(community_record_links_factory,
                                  original_links_factory=nr_links_factory),

        # not used really
        item_route='/all/not-used-but-must-be-present',
        create_permission_factory_imp=deny_all,
        delete_permission_factory_imp=deny_all,
        update_permission_factory_imp=deny_all,
        read_permission_factory_imp=check_elasticsearch,
        record_serializers={
            'application/json': 'oarepo_validate:json_response',
        },
        # search_factory_imp='restoration.objects.search:objects_search_factory'
        # default search_factory_imp: invenio_records_rest.query.default_search_factory
    )
}

# TODO: dodělat facety a filtry pro souhrnný index

RECORDS_REST_FACETS = {
    all_index_name: {
        "aggs": translate_facets({**FACETS, **CURATOR_FACETS, **DRAFT_IMPORTANT_FACETS},
                                 label='{facet_key}',
                                 value='{value_key}'),
        "filters": {**FILTERS, **CURATOR_FILTERS, **DRAFT_IMPORTANT_FILTERS}
    },
}

RECORDS_REST_SORT_OPTIONS = {
    all_index_name: {
        'alphabetical': {
            'title': 'alphabetical',
            'fields': [
                'title.cs.raw'
            ],
            'default_order': 'asc',
            'order': 1
        },
        'best_match': {
            'title': 'Best match',
            'fields': ['_score'],
            'default_order': 'desc',
            'order': 1,
        },
        'by_id': {
            'title': 'Chronological',
            'fields': ['control_number'],
            'default_order': 'desc',
            'order': 1,
        }
    }
}

RECORDS_REST_DEFAULT_SORT = {
    all_index_name: {
        'query': 'best_match',
        'noquery': 'best_match'
    }
}



from elasticsearch_dsl.query import Term, Bool
from nr_common.search import NRRecordsSearch


class AllRecordsSearch(NRRecordsSearch):
    LIST_SOURCE_FIELDS = [
        'control_number', 'oarepo:validity.valid', 'oarepo:draft', 'title',
        'dateIssued', 'creator', 'creators', 'resource_type', 'contributors', 'keywords',
        'subject', 'abstract', 'state', 'accessRights', '_files',
        'languages', 'id', '_primary_community', 'communities',
        '_administration.primaryCommunity', 'publication_date',
        '_administration.communities', 'rights',
        '$schema'
    ]


class AllRecordsDraftSearch(AllRecordsSearch):
    class ActualMeta(NRRecordsSearch.ActualMeta):
        @classmethod
        def default_filter_factory(cls, search=None, **kwargs):
            qs = NRRecordsSearch.Meta.default_filter_factory(search=search, **kwargs)
            return Bool(must=[
                qs,
                Term(**{'oarepo:draft': True})
            ])


class AllRecordsPublishedSearch(AllRecordsSearch):
    class ActualMeta(NRRecordsSearch.ActualMeta):
        @classmethod
        def default_filter_factory(cls, search=None, **kwargs):
            qs = NRRecordsSearch.Meta.default_filter_factory(search=search, **kwargs)
            return Bool(must=[
                qs,
                Term(**{'oarepo:draft': False})
            ])



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nr_all']

package_data = \
{'': ['*'],
 'nr_all': ['translations/cs/LC_MESSAGES/*', 'translations/en/LC_MESSAGES/*']}

install_requires = \
['techlib-nr-Nresults>=1.0.0a13,<2.0.0',
 'techlib-nr-common>=3.0.0a45,<4.0.0',
 'techlib-nr-datasets>=1.0.0-alpha.3,<2.0.0',
 'techlib-nr-events>=1.0.0a,<2.0.0',
 'techlib-nr-generic>=1.0.0a2,<2.0.0',
 'techlib-nr-theses>=1.0.0a3,<2.0.0']

entry_points = \
{'invenio_base.api_apps': ['nr_all = nr_all.ext:NRAll'],
 'invenio_base.apps': ['nr_all = nr_all.ext:NRAll'],
 'invenio_config.module': ['nr_all = nr_all.config'],
 'invenio_i18n.translations': ['nr_all = nr_all'],
 'invenio_pidstore.fetchers': ['nr_all = nr_all.fetchers:nr_all_id_fetcher'],
 'invenio_pidstore.minters': ['nr_all = nr_all.minters:nr_all_id_minter']}

setup_kwargs = {
    'name': 'techlib-nr-all',
    'version': '1.0.0a16',
    'description': 'Czech National Repository index that store all records from all metadata models',
    'long_description': '# nr-all\nCzech National Repository index that store all records from all metadata models.\n',
    'author': 'Daniel Kopecký',
    'author_email': 'Daniel.Kopecky@techlib.cz',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from . import config


class NRAll:
    def __init__(self, app=None, db=None):
        self.init_app(app, db)

    def init_app(self, app, db):
        self.init_config(app, db)

    def init_config(self, app, db):
        app.config.setdefault('RECORDS_REST_ENDPOINTS', {}).update(
            config.RECORDS_REST_ENDPOINTS
        )
        # TODO: dodělat facety a filtry pro souhrný index
        # app.config.setdefault('RECORDS_REST_FACETS', {}).update(
        #     config.RECORDS_REST_FACETS
        # )
